==> mi-arcfs-1.1.1.tar.gz: mi-arcfs-1.1.1/mblock.c <==
/*

    Copyright (C) 1999 Masanao Izumo <mo@goice.co.jp>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/malloc.h>
#include "mblock.h"

static MBlockNode *free_mblock_list = NULL;
#define ADDRALIGN 8
#ifndef offsetof
/* Offset of member MEMBER in a struct of type TYPE.  */
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

void mblock_init(MBlock *mblock)
{
    *mblock = NULL;
}

static MBlockNode *new_mblock_node(size_t n)
{
    MBlockNode *p;
    int offset;

    /* p->buffer + p->offset must be multipule of ADDRALIGN */
    offset = (offsetof(MBlockNode, buffer) & (ADDRALIGN - 1));
    if(offset)
	offset = ADDRALIGN - offset;
    n += offset;

    if(n <= BASE_MBLOCK_SIZE)
    {
	if(free_mblock_list)
	{
	    p = free_mblock_list;
	    free_mblock_list = free_mblock_list->next;
	}
	else
	{
	    p = (MBlockNode *)kmalloc(BASE_MBLOCK_SIZE + sizeof(MBlockNode),
				      GFP_KERNEL);
	    p->alloc_type = MBLOCK_KALLOC;
	    p->block_size = BASE_MBLOCK_SIZE;
	}
    }
    else
    {
	p = (MBlockNode *)vmalloc(n + sizeof(MBlockNode));
	p->alloc_type = MBLOCK_VALLOC;
	p->block_size = n;
    }

    p->offset = offset;
    p->next = NULL;
    return p;
}

static int enough_block_memory(MBlockNode *node, size_t n)
{
    size_t newoffset;

    if(node == NULL)
	return 0;

    newoffset = node->offset + n;

    if(newoffset < node->offset) /* exceed representable in size_t */
	return 0;

    if(newoffset > node->block_size)
	return 0;

    return 1;
}

void *mblock_alloc(MBlock *mblock, size_t nbytes)
{
    MBlockNode *p;
    void *addr;

    /* round up to ADDRALIGN */
    nbytes = ((nbytes + ADDRALIGN - 1) & ~(ADDRALIGN - 1));

    if(!enough_block_memory(*mblock, nbytes))
    {
	p = new_mblock_node(nbytes);
	p->next = *mblock;
	*mblock = p;
    }
    else
	p = *mblock;

    addr = (void *)(p->buffer + p->offset);
    p->offset += nbytes;

    return addr;
}

static void free_node(MBlockNode *p)
{
  switch(p->alloc_type)
    {
    case MBLOCK_PAGE:
      free_page((unsigned long)p);
      break;
    case MBLOCK_KALLOC:
      kfree(p);
      break;
    case MBLOCK_VALLOC:
      vfree(p);
      break;
    }
}

void mblock_reuse(MBlock *mblock)
{
    MBlockNode *p;

    if((p = *mblock) == NULL)
	return;			/* There is nothing to collect memory */

    while(p)
    {
	MBlockNode *r;

	r = p;
	p = p->next;

	if(r->block_size > BASE_MBLOCK_SIZE)
	  free_node(r);
	else
	  {
	    r->next = free_mblock_list;
	    free_mblock_list = r;
	  }
    }
    mblock_init(mblock);
}

int mblock_free(MBlock *mblock)
{
    MBlockNode *p;
    int cnt;

    p = *mblock;
    cnt = 0;
    while(p)
      {
	MBlockNode *r;

	r = p;
	p = p->next;
	free_node(r);
	cnt++;
      }
    return cnt;
}

char *mblock_strdup(MBlock *mblock, const char *str)
{
    int len;
    char *p;

    len = strlen(str);
    p = (char *)mblock_alloc(mblock, len + 1); /* for '\0' */
    memcpy(p, str, len + 1);
    return p;
}

int mblock_free_all(void)
{
  return mblock_free(&free_mblock_list);
}

==> mi-arcfs-2.2.1.tar.gz: mi-arcfs-2.2.1/mblock.c <==
/*

    Copyright (C) 1999 Masanao Izumo <mo@goice.co.jp>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include "mblock.h"

static MBlockNode *free_mblock_list = NULL;
#define ADDRALIGN 8
#ifndef offsetof
/* Offset of member MEMBER in a struct of type TYPE.  */
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

void mblock_init(MBlock *mblock)
{
    *mblock = NULL;
}

static MBlockNode *new_mblock_node(size_t n)
{
    MBlockNode *p;
    int offset;

    /* p->buffer + p->offset must be multipule of ADDRALIGN */
    offset = (offsetof(MBlockNode, buffer) & (ADDRALIGN - 1));
    if(offset)
	offset = ADDRALIGN - offset;
    n += offset;

    if(n <= BASE_MBLOCK_SIZE)
    {
	if(free_mblock_list)
	{
	    p = free_mblock_list;
	    free_mblock_list = free_mblock_list->next;
	}
	else
	{
	    p = (MBlockNode *)kmalloc(BASE_MBLOCK_SIZE + sizeof(MBlockNode),
				      GFP_KERNEL);
	    p->alloc_type = MBLOCK_KALLOC;
	    p->block_size = BASE_MBLOCK_SIZE;
	}
    }
    else
    {
	p = (MBlockNode *)vmalloc(n + sizeof(MBlockNode));
	p->alloc_type = MBLOCK_VALLOC;
	p->block_size = n;
    }

    p->offset = offset;
    p->next = NULL;
    return p;
}

static int enough_block_memory(MBlockNode *node, size_t n)
{
    size_t newoffset;

    if(node == NULL)
	return 0;

    newoffset = node->offset + n;

    if(newoffset < node->offset) /* exceed representable in size_t */
	return 0;

    if(newoffset > node->block_size)
	return 0;

    return 1;
}

void *mblock_alloc(MBlock *mblock, size_t nbytes)
{
    MBlockNode *p;
    void *addr;

    /* round up to ADDRALIGN */
    nbytes = ((nbytes + ADDRALIGN - 1) & ~(ADDRALIGN - 1));

    if(!enough_block_memory(*mblock, nbytes))
    {
	p = new_mblock_node(nbytes);
	p->next = *mblock;
	*mblock = p;
    }
    else
	p = *mblock;

    addr = (void *)(p->buffer + p->offset);
    p->offset += nbytes;

    return addr;
}

static void free_node(MBlockNode *p)
{
  switch(p->alloc_type)
    {
    case MBLOCK_PAGE:
      free_page((unsigned long)p);
      break;
    case MBLOCK_KALLOC:
      kfree(p);
      break;
    case MBLOCK_VALLOC:
      vfree(p);
      break;
    }
}

void mblock_reuse(MBlock *mblock)
{
    MBlockNode *p;

    if((p = *mblock) == NULL)
	return;			/* There is nothing to collect memory */

    while(p)
    {
	MBlockNode *r;

	r = p;
	p = p->next;

	if(r->block_size > BASE_MBLOCK_SIZE)
	  free_node(r);
	else
	  {
	    r->next = free_mblock_list;
	    free_mblock_list = r;
	  }
    }
    mblock_init(mblock);
}

int mblock_free(MBlock *mblock)
{
    MBlockNode *p;
    int cnt;

    p = *mblock;
    cnt = 0;
    while(p)
      {
	MBlockNode *r;

	r = p;
	p = p->next;
	free_node(r);
	cnt++;
      }
    return cnt;
}

char *mblock_strdup(MBlock *mblock, const char *str)
{
    int len;
    char *p;

    len = strlen(str);
    p = (char *)mblock_alloc(mblock, len + 1); /* for '\0' */
    memcpy(p, str, len + 1);
    return p;
}

int mblock_free_all(void)
{
  return mblock_free(&free_mblock_list);
}
